import urllib.request, pandas as pd, os, itertools, torch
from torch.utils.data import Dataset
from os.path import join
from pathlib import Path

data_path = "./data/"

def download_dataset(source_url, destination):
    Path(data_path).mkdir(parents=True, exist_ok=True)
    if not os.path.isfile(source_url):
        urllib.request.urlretrieve(source_url, destination)

class ContextualBanditDataset(Dataset):

    def __init__(self, contexts, target, optimal_accuracy = 1):
        """
        contexts: matrix of size N x D
        target: the list of size N of the classes associated to the contexts
        """
        assert len(contexts) == len(target)

        self._contexts = contexts
        self._target = target
        self._K = len(self._target.astype('category').cat.categories)
        self._D = len(self._contexts[0])
        
        self.optimal_accuracy = optimal_accuracy
            
    def __len__(self):
        return len(self._target)

    def __getitem__(self, idx):
        return self._contexts[idx], self._target.iloc[idx]
    
    @property
    def K(self): 
        return self._K
    
    @property
    def D(self): 
        return self._D        
        

def get_cov_dataset():
    """
    Forest Cover Type DataSet

    Elevation                               quantitative    meters                       Elevation in meters
    Aspect                                  quantitative    azimuth                      Aspect in degrees azimuth
    Slope                                   quantitative    degrees                      Slope in degrees
    Horizontal_Distance_To_Hydrology        quantitative    meters                       Horz Dist to nearest surface water features
    Vertical_Distance_To_Hydrology          quantitative    meters                       Vert Dist to nearest surface water features
    Horizontal_Distance_To_Roadways         quantitative    meters                       Horz Dist to nearest roadway
    Hillshade_9am                           quantitative    0 to 255 index               Hillshade index at 9am, summer solstice
    Hillshade_Noon                          quantitative    0 to 255 index               Hillshade index at noon, summer soltice
    Hillshade_3pm                           quantitative    0 to 255 index               Hillshade index at 3pm, summer solstice
    Horizontal_Distance_To_Fire_Points      quantitative    meters                       Horz Dist to nearest wildfire ignition points
    Wilderness_Area (4 binary columns)      qualitative     0 (absence) or 1 (presence)  Wilderness area designation
    Soil_Type (40 binary columns)           qualitative     0 (absence) or 1 (presence)  Soil Type designation
    Cover_Type (7 types)                    integer         1 to 7                       Forest Cover Type designation
    """

    cov_url = "https://archive.ics.uci.edu/ml/machine-learning-databases/covtype/covtype.data.gz"
    cov_path = join(data_path, "covtype.data.gz")

    download_dataset(cov_url, cov_path)

    names = list(itertools.chain([
        "Elevation",
        "Aspect",
        "Slope",
        "Horizontal_Distance_To_Hydrology",
        "Vertical_Distance_To_Hydrology",
        "Horizontal_Distance_To_Roadways",
        "Hillshade_9am",
        "Hillshade_Noon",
        "Hillshade_3pm",
        "Horizontal_Distance_To_Fire_Points"],
        [f"Wilderness_Area_{i}" for i in range(4)],
        [f"Soil_Type_{i}" for i in range(40)],
        ["Cover_Type"]
    ))

    df = pd.read_csv(cov_path, compression='gzip', header = None, names = names)

    """
    Following Allesiardo et al (https://arxiv.org/abs/1409.8191) we discretize each continuous features using an equal frequency binning of size 5
    """
    continuous_features = ["Elevation", "Aspect", "Slope", "Horizontal_Distance_To_Hydrology", "Vertical_Distance_To_Hydrology", "Horizontal_Distance_To_Roadways", "Hillshade_9am", "Hillshade_3pm", "Horizontal_Distance_To_Fire_Points", "Hillshade_Noon"]
    nb_bin = 5

    for i_feature_name in continuous_features:
        tmp = pd.get_dummies(pd.qcut(df[i_feature_name], nb_bin))
        tmp.columns = [f"{i_feature_name}_{i}" for i in range(nb_bin)]

        del df[i_feature_name]
        for i_col in tmp.columns:
            df[i_col] = tmp[i_col]

        del tmp

    target_name = "Cover_Type"

    target = df[target_name].astype('category').cat.codes
    del df[target_name]

    contexts = torch.from_numpy(df.values).float()
    del df
    
    """
    The optimal policy used by Allesiardo et al (https://arxiv.org/abs/1409.8191) achieves 93% of good classification
    """
    return ContextualBanditDataset(contexts, target, 0.93)



from torch.utils.data import DataLoader
from .data  import ContextualBanditDataset

class ContextualBanditGame():

    def __init__(self, dataset, T):
        
        assert isinstance(dataset, ContextualBanditDataset)
        
        self._dataset = dataset
        self.T = T
        self.init_game()
        
    @property
    def optimal_accuracy(self): 
        return self._dataset.optimal_accuracy

    def init_game(self):
        self._generator = self._get_generator()
        _ = self.get_context()    
        
    def _get_generator(self):
        """
        Return a generator yielding the contexts and updating the label
        """
        
        self._dataloader = DataLoader(self._dataset, batch_size=1,
                        shuffle=True)
        self._t = 0
        self._context = None
        self._label = None
        self._action_played = False
        
        while self._t < self.T: # We go through the dataloader only if the game is not over
            for i_context, i_label in self._dataloader:
                if self._t >= self.T: # If the game is over, the game just return None everywhere and we break the loop
                    self._context = None
                    self._label = None
                    break
                else:
                    self._action_played = False # We wait for the new action to be played
                    self._context = i_context # We update the current context and label
                    self._label = i_label
                    
                # We yield the current context, the label is not provided and will be used to compute the reward in the play function
                yield self._context
                
                while not self._action_played:# Return the same context while the next action is unplayed
                    yield self._context
                    
                self._t = self._t+1
                    
    def get_context(self):
        """
        Return the next context, None if the game is over
        """
        try:
            return next(self._generator)
        except:
            return None
    
    def play(self, action):
        """ Play an action and receive a reward. The reward is None if the game is already over """
        if self._label is None:
            return None
        
        self._action_played = True # We told the generator that an action has be played
        
        if self._label.data == action: #If the action match the label the reward is 1 else 0
            reward =  1
        else:
            reward =  0
            
        _ = self.get_context() # We go to the next steps, just in case the player is playing actions without requesting a new context
        
        return reward


import torch, numpy as np
from random import randrange

class RandomBandit():
    """
    The Random Bandit
    """
    def __init__(self, K):
        self._K = K
        
    def select(self, x = None):
        return randrange(self._K)
    
    def observe(self, k, x, y):
        """
        Does nothing, is here just for compatibility of the API
        """
        pass


class Banditron():
    """
    Implementation of the Banditron paper (https://www.cse.huji.ac.il/~shais/papers/TewariShKa08.pdf)
    """
    def __init__(self, action_count, context_dimension, gamma = 0.05):
        
        self._K = action_count
        self._D = context_dimension
        self._gamma = gamma
        
        self._weights = torch.zeros(context_dimension, action_count)

    def select(self, context):
        with torch.no_grad():
            XW =  context.mm(self._weights)
            _, self.best_actions = torch.max(XW, 1)

            mask = torch.zeros(XW.shape).scatter (1, self.best_actions.unsqueeze(1), 1.0)
            self._probabilities = (1 - self._gamma) * mask + torch.zeros((len(XW), self._K)) + self._gamma/self._K

            k = [np.random.choice(self._K, p=p.data.numpy()/sum(p.data.numpy())) for p in self._probabilities]

            return k[0]
    
    def observe(self, played_action, context, reward):
        """
        Doesn't support batches of rewards
        """
        with torch.no_grad():
            self.update = context * ((reward - 1)/self._probabilities[0][played_action] - 1)
            self._weights[:,played_action:played_action+1] += self.update.squeeze(0).unsqueeze(1)
            
            
class LinUCB():
    """
    Implementation of the LinUCB algorithm with dijoint linear model (https://arxiv.org/pdf/1003.0146.pdf)
    """
    
    def __init__(self, K, D, inversion_interval = 100):
        """
        The inversion of the matrix is an expensive task and will be performed only every "inversion_interval" iterations
        """

        self._K = K
        self._D = D
        self._inversion_interval = inversion_interval
        self.reset()

    def select(self, context):
        """
        Selection the most optimal action using an upper confidence bound
        """
        with torch.no_grad():
            value = torch.Tensor([context.mm(i_theta) for i_theta in self._theta])
            confidence_interval = torch.stack([torch.sqrt(context.mm(self._A_inv[k]).mm(context.transpose(0,1))) for k in range(self._K)]).reshape(len(context), self._K)
            _, best_action = torch.max(value + confidence_interval, 1)
            return int(best_action[0])

    def observe(self, played_arm, context, reward, update = False):
        """
        Update of the counts. Predictions will be updated only after calling self._invert()
        """
        with torch.no_grad():
            self._A[played_arm] = self._A[played_arm] + context.transpose(0,1).mm(context)
            self._b[played_arm] = self._b[played_arm] + torch.transpose(context * reward, 0, 1)
            if update or ((self.t+1)%self._inversion_interval == 0):
                self._invert()

            self.t += 1

    def reset(self):
        
        self.t = 0
        self._A = [torch.eye(self._D) for k in range(self._K)]
        self._b = [torch.zeros((self._D, 1)) for k in range(self._K)]

        self._invert()

    def _invert(self):
        """
        Matrix inversion to update the model
        """
        with torch.no_grad():
            self._A_inv = [a.inverse() for a in self._A]
            self._theta = [self._A_inv[k].mm(self._b[k]) for k in range(self._K)]



import itertools, numpy as np, torch, torch.nn as nn

def get_simple_network(context_dimension, layer_count, layer_size):
    """
    Simple Network used in the NeuralBandit paper (https://hal.archives-ouvertes.fr/hal-01117311/document)
    """
    last_dim = context_dimension
    layers = []

    for i in range(layer_count):
        layers.append(nn.Linear(last_dim, layer_size))
        layers.append(nn.Sigmoid())
        last_dim = layer_size

    layers.append(nn.Linear(last_dim, 1))
    layers.append(nn.Sigmoid())

    return nn.Sequential(*layers)
    

class NeuralBandit():
    """
    Implementation of the NeuralBandit paper (https://hal.archives-ouvertes.fr/hal-01117311/document)
    """
    def __init__(self, action_count, context_dimension, layer_count = 0, layer_size = 0, lr = 0.1, gamma = 0.05, get_network = get_simple_network):
        
        self._K = action_count
        self._D = context_dimension
        self._gamma = gamma
        
        self._nn = [get_network(context_dimension, layer_count, layer_size) for k in range(self._K)]
        self._opt = torch.optim.SGD(list(itertools.chain(*[list(i_nn.parameters()) for i_nn in self._nn])), lr=lr, momentum=0.9)
        self._loss = torch.nn.MSELoss()
                
    def select(self, context):
        
        self._scores =  torch.stack([i_nn(context) for i_nn in self._nn]).reshape(len(context), self._K)
        _, self.best_actions = torch.max(self._scores, 1)

        mask = torch.zeros(self._scores.size()).scatter (1, self.best_actions.unsqueeze(1), 1.0)
        self._probabilities = (1 - self._gamma) * mask + torch.zeros((len(self._scores), self._K)) + self._gamma/self._K

        k = [np.random.choice(self._K, p=p.data.numpy()/sum(p.data.numpy())) for p in self._probabilities]

        return k[0]
    
    def observe(self, played_action, context, reward):
        """
        Doesn't support batches of rewards
        """
        loss = self._loss(self._scores[0][played_action], torch.tensor(reward).float())
        self._opt.zero_grad()
        loss.backward()
        self._opt.step()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neuralbandit']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18.1,<2.0.0', 'pandas>=0.25.3,<0.26.0', 'torch>=1.4.0,<2.0.0']

setup_kwargs = {
    'name': 'neuralbandit',
    'version': '0.1.2',
    'description': 'Reproducing the experiments of the NeuralBandit and BanditForest papers. ',
    'long_description': '# NeuralBandit\n\nThis WIP repository reproduces the experiments of the [NeuralBandit](https://hal.archives-ouvertes.fr/hal-01117311/document) and [BanditForest](http://proceedings.mlr.press/v51/feraud16.html) papers. The code of BanditForest is available [here](https://www.researchgate.net/publication/308305599_Test_code_for_Bandit_Forest_algorithm).\n\n### Table of contents\n- [Steps to reproduce](#steps-to-reproduce)\n- [Citations](#citations)\n    + [NeuralBandit](#neuralbandit-1)\n    + [BanditForest](#banditforest)\n\n### TODO List\n* Doing the requirements.txt\n* Automating the run of experiments\n* NeuralBandit Paper\n    * Implementation of NeuralBandit.A and .B\n    * Adding concept drift to the Game\n\n* BanditForest Paper\n    * Adding Census et Adult dataset\n    * Adding Noise to the reward\n\n### Steps to reproduce\n```python\n\nfrom neuralbandit import get_cov_dataset, ContextualBanditGame, NeuralBandit\nfrom neuralbandit.sota import RandomBandit, BanditTron, LinUCB\n\ndataset = get_cov_dataset()\n\ncumulative_reward = 0\nT = int(2.5E6)\n\ngame = ContextualBanditGame(dataset, T)\n#player = RandomBandit(dataset.K)\n#player = LinUCB(dataset.K, dataset.D)\nplayer = NeuralBandit(dataset.K, dataset.D, layer_count = 1, layer_size = 64, gamma = 0.05)\n\nfor t in tqdm(range(T)):\n        context = game.get_context()\n        action = player.select(context)\n        reward = game.play(action)\n        \n        cumulative_reward += reward\n        \n        player.observe(action, context, reward)\n\ncumulative_regret = (T*game.optimal_accuracy - cumulative_reward)\nprint(cumulative_regret)\n```\n\n### Citations\n#### NeuralBandit\n```\n@InProceedings{10.1007/978-3-319-12637-1_47,\nauthor="Allesiardo, Robin\nand F{\\\'e}raud, Rapha{\\"e}l\nand Bouneffouf, Djallel",\neditor="Loo, Chu Kiong\nand Yap, Keem Siah\nand Wong, Kok Wai\nand Teoh, Andrew\nand Huang, Kaizhu",\ntitle="A Neural Networks Committee for the Contextual Bandit Problem",\nbooktitle="Neural Information Processing",\nyear="2014",\npublisher="Springer International Publishing",\naddress="Cham",\npages="374--381",\nabstract="This paper presents a new contextual bandit algorithm, NeuralBandit, which does not need hypothesis on stationarity of contexts and rewards. Several neural networks are trained to modelize the value of rewards knowing the context. Two variants, based on multi-experts approach, are proposed to choose online the parameters of multi-layer perceptrons. The proposed algorithms are successfully tested on a large dataset with and without stationarity of rewards.",\nisbn="978-3-319-12637-1"\n}\n```\n#### BanditForest\n```\n@InProceedings{pmlr-v51-feraud16,\n  title = \t {Random Forest for the Contextual Bandit Problem},\n  author = \t {Raphaël Féraud and Robin Allesiardo and Tanguy Urvoy and Fabrice Clérot},\n  booktitle = \t {Proceedings of the 19th International Conference on Artificial Intelligence and Statistics},\n  pages = \t {93--101},\n  year = \t {2016},\n  editor = \t {Arthur Gretton and Christian C. Robert},\n  volume = \t {51},\n  series = \t {Proceedings of Machine Learning Research},\n  address = \t {Cadiz, Spain},\n  month = \t {09--11 May},\n  publisher = \t {PMLR},\n  pdf = \t {http://proceedings.mlr.press/v51/feraud16.pdf},\n  url = \t {http://proceedings.mlr.press/v51/feraud16.html},\n  abstract = \t {To address the contextual bandit problem, we propose an online random forest algorithm. The analysis of the proposed algorithm is based on the sample complexity needed to find the optimal decision stump. Then, the decision stumps are recursively stacked in a random collection of decision trees, BANDIT FOREST. We show that the proposed algorithm is optimal up to logarithmic factors. The dependence of the sample complexity upon the number of contextual variables is logarithmic. The computational cost of the proposed algorithm with respect to the time horizon is linear. These analytical results allow the proposed algorithm to be efficient in real applications , where the number of events to process is huge, and where we expect that some contextual variables, chosen from a large set, have potentially non-linear dependencies with the rewards. In the experiments done to illustrate the theoretical analysis, BANDIT FOREST obtain promising results in comparison with state-of-the-art algorithms.}\n}\n\n```\n',
    'author': 'rallesiardo',
    'author_email': 'robin.allesiardo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rallesiardo/NeuralBandit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
